(** Basic machine model *)

(** operational semantics *)

(** Chunxiao Lin @ 2006 *)

Require Export state.
Require Export natutil.
Require Export basicutil.

Require Export List.
Require Import Omega.

Set Implicit Arguments.
Unset Strict Implicit.

Notation WSIZE := 4.

(** (cdhp, hp, rf, iseq) *)

(** ******************************************************************* *)
(** istructions *)
(** ******************************************************************* *)
Inductive comm : Set :=
  | andi1 : reg -> reg -> comm
  | andi7 : reg -> reg -> comm
  | srl1 : reg -> reg -> comm
  | addu : reg -> reg -> reg -> comm
  | addiu : reg -> reg -> nat -> comm
  | subiu : reg -> reg -> nat -> comm
  | subu : reg -> reg -> reg -> comm
  | sltu : reg -> reg -> reg -> comm
  | lw : reg -> nat -> reg -> comm
  | sw : reg -> nat -> reg -> comm.

Inductive iseq : Set :=
  | seq : comm -> iseq -> iseq
  | br : btype -> reg -> reg -> nat -> iseq -> iseq
  | j : nat -> iseq
  | jal : nat -> nat -> iseq
  | jr : reg -> iseq.

Definition beq := br bteq.
Definition bne := br btne.
Definition blt := br btlt.
Definition bge := br btge.


(** ******************************************************************* *)
(** program *)
(** ******************************************************************* *)
Definition code_heap : Set := nat -> option iseq.
  
Definition prog : Set := (prod (prod code_heap state) iseq).
  
Definition is_of (p : prog) := snd p.
Definition ch_of (p : prog) := fst (fst p).
Definition st_of (p : prog) := snd (fst p).


(** ******************************************************************* *)
(** op sem *)
(** ******************************************************************* *)
Definition cstep (c : comm) (s : state) : option state :=
  match c with
    | andi1 rd rs =>
      Some (rset s rd (if bodd (rget s rs) then 1 else 0))
    | andi7 rd rs =>
      Some (rset s rd (mod8 (rget s rs)))
    | srl1 rd rs =>
      Some (rset s rd (div2 (rget s rs)))
    | addu rd rs rt => 
      Some (rset s rd (rget s rs + rget s rt))
    | addiu rt rs imm => 
      Some (rset s rt (rget s rs + imm))
    | subiu rt rs imm => 
      Some (rset s rt (rget s rs - imm))
    | subu rd rs rt => 
      Some (rset s rd (rget s rs - rget s rt))
    | sltu rd rs rt =>
      Some (rset s rd (if blt_nat (rget s rs) (rget s rt) then 1 else 0))
    | lw rt offset rs => 
      match (hget s (rget s rs + offset)) with
	| Some v => Some (rset s rt v)
	| None => None
	end
    | sw rt offset rs => 
      match (hget s (rget s rs + offset)) with
	| Some v => 
	  Some (hset s (rget s rs + offset) (rget s rt))
	| None => None
      end
  end.

Definition ststep (s : state) (is : iseq) : option state:=
  match is with 
    | seq c is' => cstep c s
    | jal fret target => Some (rset s r31 fret)
    | _ => Some s
  end.

Definition iseqstep (ch : code_heap) (s : state) (is : iseq) : option iseq :=
  match is with 
    | seq c is' => Some is'
    | br bt rs rt offset is' => 
      if (btest bt (rget s rs) (rget s rt)) then (ch offset) else (Some is')
    | jr rs => ch (rget s rs)
      | j target => ch target
    | jal fret target => ch target
    end.

Definition step (p : prog) : option prog :=
  match ststep (st_of p) (is_of p), iseqstep (ch_of p) (st_of p) (is_of p) with
    | Some s', Some is' => Some ((ch_of p, s'), is')
    | _, _ => None
  end.

Fixpoint stepn (n : nat) (p : prog) {struct n} : option prog :=
  match n with
      | O => Some p
    | S n' =>
      match step p with
	| Some p' => stepn n' p'
	| None => None
      end
  end.


(** ******************************************************************* *)
(** others *)
(** ******************************************************************* *)
Notation "i ; I" := (seq i I) (at level 30, right associativity).
Notation "i ;; I" := (i (I:iseq)) (at level 30, right associativity).

Ltac dst :=
  match goal with
    | [|- context [(ch_of (?c, ?s, ?i))]] => 
      replace (ch_of (c, s, i)) with c; [ dst | reflexivity ]
    | [|- context [(st_of (?c, ?s, ?i))]] => 
      replace (st_of (c, s, i)) with s; [ dst | reflexivity ]
    | [|- context [(is_of (?c, ?s, ?i))]] => 
      replace (is_of (c, s, i)) with i; [ dst | reflexivity ]
    | [|- context [(hp_of (?r, ?h))]] => 
      replace (hp_of (r, h)) with h; [ dst | reflexivity ]
    | [|- context [(rf_of (?r, ?h))]] => 
      replace (rf_of (r, h)) with r; [ dst | reflexivity ]
    | _  => idtac
  end.

Lemma ch_lab_dec : forall (ch : code_heap) lab, 
  (exists is, ch lab = Some is) \/ (ch lab = None).
Proof.
  intros.
  destruct (ch lab); [left; exists i | right]; reflexivity.
Qed.
