(** GC Safety for TAL *)

(** Reachability *)

(** Chunxiao Lin @ 2006 *)


Set Implicit Arguments.
Unset Strict Implicit.

Require Import machine2.

(** ******************************************************************* *)
(** border of the object heap *)
(** ******************************************************************* *)
Variable ohpst : nat.

Variable ohplen : nat.

Definition ost := (8 * (S ohpst)).

Definition oed := (8 * (S ohpst + (S ohplen))).

Lemma ost_mod8 : mod8 ost = 0.
Proof.
  unfold ost.
  apply mod8_mul_8.
Qed.

Lemma oed_mod8 : mod8 oed = 0.
Proof.
  unfold oed.
  apply mod8_mul_8.
Qed.


(** ******************************************************************* *)
(** valid heap pointer *)
(** ******************************************************************* *)
Notation align8 := (fun x => mod8 x = 0).

Definition vptr x := (ost <= x < oed) /\ align8 x.

(** ******************************************************************* *)
(** computational valid heap pointer *)
(** ******************************************************************* *)
Notation align8b := (fun x => beq_nat (mod8 x) 0).

Definition vptrb x := (blt_nat x oed) && (negb (blt_nat x ost)) && align8b x.

(* lemmas: vptrb *)
Lemma vptrb_intro : forall x, vptr x -> vptrb x = true.
Proof.
  unfold vptr; unfold vptrb; intros x [H1 H2].
  splitb; try (rewrite H2); firstorder.
Qed.

Lemma vptrb_elim : forall x, vptrb x = true -> vptr x.
Proof.
  unfold vptr; unfold vptrb; intros.
  elimbool.
  hypreplace B0 (beq_nat_eq  _ _ (sym_eq B0)).
  hypreplace B1 (blt_lt B1).
  hypreplace B2 (blt_ge (negb_elim1 B2)).
  repeat split; firstorder.
Qed.

Lemma vptrb_intro_false : forall x, (~ vptr x) -> vptrb x = false.
Proof.
  intros.
  destruct (bool_dec (vptrb x)); trivial.
  destruct H.
  apply (vptrb_elim H0).
Qed.

Lemma vptrb_elim_false : forall x, vptrb x = false -> (~ vptr x).
Proof.
  intros.
  destruct (bool_dec (vptrb x)); intro F.
  rewrite H in H0; discriminate.
  rewrite (vptrb_intro F) in H0; discriminate.
Qed.

Hint Resolve vptrb_intro vptrb_intro_false vptrb_elim vptrb_elim_false : vptr.

Lemma vptr_dec : forall x,
  vptr x \/ ~ vptr x.
Proof.
  intros.
  destruct (bool_dec (vptrb x));
    [left | right]; auto with vptr.
Qed.


(** ******************************************************************* *)
(** inductive reachability *)
(** ******************************************************************* *)
Inductive reach : heap -> nat -> nat -> Prop :=
  | rch_refl : forall hp x, vptr x -> reach hp x x
  | rch_next : forall hp x y z, 
    (heap_get hp x = Some z \/ heap_get hp (x + WSIZE) = Some z) ->
    reach hp z y -> 
    vptr x -> vptr y ->
    reach hp x y.

(** lemmas: reach *)
Lemma reach_vptrx : forall x y hp,
  reach hp x y -> vptr x.
Proof.
  intros.
  inversion_clear H; trivial.
Qed.

Lemma reach_vptry : forall x y hp,
  reach hp x y -> vptr y.
Proof.
  intros.
  inversion_clear H; trivial.
Qed.

Lemma reach_trans : forall x y z hp, 
  reach hp x y -> reach hp y z -> reach hp x z.
Proof.
  intros x y z hp Hxy Hyz.
  induction Hxy.
  trivial.
  hypreplace Hyz (IHHxy Hyz).
  apply rch_next with z0; trivial.
  apply (reach_vptry Hyz).
Qed.

Lemma rch_next2 : forall h x y z,
  vptr y ->
  reach h x z ->
  (heap_get h z = Some y \/ heap_get h (z + WSIZE) = Some y) ->
  reach h x y.
Proof.
  intros.
  induction H0.
  apply rch_next with y; trivial.
  constructor; trivial.
  hypreplace H1 (IHreach H1).
  assert (Hp := reach_vptrx H1).
  apply reach_trans with z; trivial.
  apply rch_next with z; trivial.
  constructor; trivial.
Qed.

Lemma heap_remove_reach : forall h x y l,
  reach (heap_remove h l) x y -> reach h x y.
Proof.
  intros.
  assert (exists h', h' = heap_remove h l).
  exists (heap_remove h l).
  reflexivity.
  destruct H0 as [h' Hh'].
  rewrite <- Hh' in H.
  induction H.
  constructor; trivial.
  subst.
  apply rch_next with z; trivial.
  destruct H; [left | right]; generalize H; clear H.
  compare x l; intro e; subst.
  rewrite heap_remove_get_eq; intro; discriminate.
  rewrite heap_remove_get_ne; trivial.
  compare (x + 4) l; intro e; subst.
  rewrite heap_remove_get_eq; intro; discriminate.
  rewrite heap_remove_get_ne; trivial.
  firstorder.
Qed.

(** ******************************************************************* *)
(** representing reachability with heap path *)
(** ******************************************************************* *)
Fixpoint heap_path (h : heap) (path : list nat) {struct path}: Prop :=
  match path with
    | a :: path' => vptr a /\
      match path' with
	| a' :: path'' => 
	  (heap_get h a = Some a' \/ 
	    heap_get h (a + WSIZE) = Some a') /\ 
	  heap_path h path'
	| nil => True
      end
    | nil => True
  end.
      
(** lemmas: heap path *)
Lemma heap_path_next : forall h x p,
  heap_path h (x :: p) -> heap_path h p.
Proof.
  simpl; intros.
  destruct p;
    firstorder.
Qed.

Lemma heap_path_vptr : forall p h x,
  heap_path h p -> In x p -> vptr x.
Proof.
  induction p; intros.
  inversion_clear H0.
  simpl in H0; destruct H0; subst.
  firstorder.
  hypreplace H (heap_path_next H).
  firstorder.
Qed.

Lemma heap_path_elim : forall p x y h,
  heap_path h p -> head p = Some x -> last p = Some y -> reach h x y.
Proof.
  induction p; intros.
  discriminate.
  simpl in H0.
  injection H0; intros; subst.
  destruct p.
  simpl in * .
  injection H1; intros; subst.
  constructor; firstorder.
  rewrite last_next in H1.
  assert (Hp := heap_path_next H).
  destruct H as [Hptr [Hget _]].
  apply rch_next with n; trivial.
  firstorder.
  apply (heap_path_vptr Hp (last_elim H1)).
Qed.

Lemma heap_path_intro : forall x y h,
  reach h x y -> (exists p, heap_path h p /\ head p = Some x /\ last p = Some y).
Proof.
  intros.
  induction H.
  exists (x::nil).
  firstorder.
  destruct IHreach as [p' [Hp [Hz Hy]]].
  exists (x :: p').
  split.
  simpl.
  destruct p'.
  firstorder.
  injection Hz; intros; subst.
  firstorder.
  split.
  reflexivity.
  destruct p'. 
  discriminate.
  rewrite last_next; trivial.
Qed.

Lemma heap_path_eqv_heap' : forall p h h',
  heap_path h p -> p <> nil ->
  (forall l, In l p -> heap_get h l = heap_get h' l /\
    heap_get h (l + WSIZE) = heap_get h' (l + WSIZE)) ->
  heap_path h' p.
Proof.
  induction p; intros.
  destruct H0; reflexivity.
  destruct p.
  trivial.
  destruct H as [Hv [Hval Hp]].
  split; trivial.
  split.
  destruct (H1 a) as [V1 V2].
  intuition.
  destruct Hval; [left; rewrite <- V1 | right; rewrite <- V2];
    trivial.
  apply IHp with h; trivial.
  intro F; discriminate.
  intros; apply H1; right; trivial.
Qed.

Lemma heap_path_eqv_heap : forall p h h' x y,
  heap_path h p -> head p = Some x -> last p = Some y ->
  (forall l, In l p -> heap_get h l = heap_get h' l /\
    heap_get h (l + WSIZE) = heap_get h' (l + WSIZE)) ->
  heap_path h' p /\ head p = Some x /\ last p = Some y.
Proof.
  induction p; intros.
  discriminate.
  destruct p.
  simpl in * .
  injection H0; injection H1; intros; subst; subst.
  firstorder.
  destruct (IHp h h' n y) as [X1 [X2 X3]]; trivial.
  destruct H; firstorder.
  intros; apply H2; right; trivial.
  split.
  split.
  destruct H; trivial.
  split.
  destruct H as [_ [H _]].
  destruct (H2 a).
  firstorder.
  destruct H; [left; rewrite <- H3 
    | right; rewrite <- H4]; trivial.
  destruct H; firstorder.
  split; trivial.
Qed.

Lemma heap_path_break : forall p h p1 p2,
  p = p1 ++ p2 -> heap_path h p -> heap_path h p1 /\ heap_path h p2.
Proof.
  induction p; intros.
  destruct p1; destruct p2;
    try discriminate.
  firstorder.
  destruct p1.
  simpl in H; subst; firstorder.
  rewrite <- app_comm_cons in H.
  injection H; intros; clear H; subst.
  assert (H := heap_path_next H0).
  destruct (IHp _ _ _ (refl_equal (p1 ++ p2)) H).
  split; trivial.
  destruct p1.
  destruct H0; firstorder.
  rewrite <- app_comm_cons in H0.
  destruct H0 as [X1 [X2 _]].
  repeat (split; trivial).
Qed.

Lemma heap_path_in_reach : forall p h x y,
  heap_path h p -> In y p -> head p = Some x ->
  reach h x y.
Proof.
  induction p; intros.
  discriminate.
  injection H1; intros; subst.
  simpl in H0; destruct H0; subst.
  constructor; destruct H; trivial.
  destruct p.
  destruct H0.
  destruct H as [Hx [Hxv Hp]].
  apply rch_next with n; trivial.
  apply IHp; trivial.
  apply heap_path_vptr with (n :: p) h; assumption.
Qed.

Lemma reach_heap_eqv_reach : forall h h' a x,
  (forall y, reach h a y -> 
    heap_get h y = heap_get h' y /\
    heap_get h (y + 4) = heap_get h' (y + 4)) ->
  reach h a x -> reach h' a x.
Proof.
  intros.
  destruct (heap_path_intro H0) as [p [Hp [Ha Hy]]].
  apply heap_path_elim with p; trivial.
  apply heap_path_eqv_heap' with h; trivial.
  destruct p; try (intro F); discriminate.
  intros.
  apply H.
  apply (heap_path_in_reach Hp H1 Ha).
Qed.

(** remove a loop does not affect reachability *)
Lemma reach_heap_remove0_reach : forall h l x y,
  vptr l ->
  heap_get h l = Some x ->
  reach h x y -> reach (heap_remove h l) x y.
Proof.
  intros.
  destruct (heap_path_intro H1) as [p [Hp [Hx Hy]]].
  destruct (natlist_in_dec p l) as [IN | Nin].

  (* case 1: l \in p *)
  destruct (natlist_break IN) as [p1 [p2 [Happ Nin]]].
  destruct (heap_path_break Happ Hp). 

  destruct p2.

  (* case 1.1: p = p1 ++ l :: nil, l = y *)
  subst.
  assert (E := Hy).
  rewrite last_app in E; injection E; intros; subst.
  clear H2 H3 Nin E.
  destruct (natlist_break2 IN) as [p2 [p3 [Happ NIin]]].
  destruct (heap_path_break Happ Hp).
  destruct (list_last_dec p2) as [| [pt [t Ht]]]; subst.
  rewrite Happ in Hx; injection Hx; intros; subst.
  constructor; trivial.
  apply rch_next2 with t; trivial.
  destruct (heap_path_eqv_heap (h := h) (h' := heap_remove h y) 
    (p := (pt ++ t :: nil)) (x:= x) (y:=t)); trivial.
  rewrite Happ in Hx; rewrite head_app_not_nil in Hx; trivial.
  destruct pt; intro F; discriminate.
  rewrite last_app; reflexivity.
  intros.
  assert (e : l <> y).
  intro F; subst; tauto.
  repeat (rewrite heap_remove_get_ne; trivial).
  firstorder.
  intro F.
  hypreplace H4 (heap_path_vptr H2 H4).
  destruct H4; destruct H.
  subst.
  rewrite (mod8_plus4 H5) in H6.
  discriminate.
  apply heap_path_elim with (pt ++ t :: nil); firstorder.

  assert (e : t <> y).
  intro F; subst; apply NIin.
  apply in_or_app.
  firstorder.
  rewrite app_x_nil_app in Happ.
  destruct (heap_path_break Happ Hp).
  destruct H5 as [Hv [Ht _]].
  destruct Ht; [left | right];
    rewrite heap_remove_get_ne; trivial.
  intro F.
  destruct Hv; destruct H.
  subst.
  rewrite (mod8_plus4 H7) in H8.
  discriminate.

  (* case 1.2: p = p1 ++ l :: n :: p2 *)
  assert (Hp' := H3).
  destruct H3 as [Hl [Hv _]].
  destruct Hv.

  (* case 1.2.1: n = x *)
  rewrite H3 in H0; injection H0; intros; subst.
  assert (Happ := app_x_nil_app p1 (x ::p2) l). 
  destruct (heap_path_break (sym_eq Happ) Hp).
  destruct (heap_path_eqv_heap (h := h) (h' := heap_remove h l) 
    (p := (x :: p2)) (x:= x) (y:=y)); trivial.
  rewrite <- Happ in Hy; rewrite last_app in Hy; trivial.
  intros.
  assert (e : l0 <> l).
  intro F; subst; tauto.
  repeat (rewrite heap_remove_get_ne; trivial).
  firstorder.
  intro F.
  hypreplace H6 (heap_path_vptr H5 H6).
  destruct H6; destruct H.
  subst.
  rewrite (mod8_plus4 H7) in H8.
  discriminate.
  apply heap_path_elim with (x :: p2); firstorder.

  (* case 1.2.2: h(l + 4) = n *)
  apply reach_trans with l.
  destruct (natlist_in_dec p1 l).

  (* case 1.2.2.1: l \in p1 *)
  destruct (natlist_break2 H4) as [p3 [p4 [Happ' Nin']]].
  destruct (heap_path_break Happ Hp) as [Hp1 _].
  
  destruct (list_last_dec p3) as [| [p5 [z Hz]]].

  (* case 1.2.2.1.1:  p1 = nil :: l :: p4 *)
  subst p3; simpl in Hx.
  rewrite Happ in Hx; rewrite Happ' in Hx.
  injection Hx; intros; subst x; constructor; trivial.

  (* case 1.2.2.1.2:  p1 = (p5 ++ z :: nil) ++ l :: p4 *)
  apply rch_next2 with z; trivial.

  destruct (heap_path_break Happ' Hp1) as [Hp3 _]; trivial.
  destruct (heap_path_eqv_heap (h := h) (h' := heap_remove h l) 
    (p := p3) (x:= x) (y:=z)); trivial.
  subst; rewrite head_app_not_nil in Hx; trivial.
  rewrite head_app_not_nil in Hx; trivial.
  destruct p5; intro F; discriminate.
  destruct p5; intro F; discriminate.
  subst p3; rewrite last_app; reflexivity.
  intros.
  assert (e : l0 <> l).
  intro F; subst; tauto.
  repeat (rewrite heap_remove_get_ne; trivial).
  firstorder.
  intro F.
  hypreplace H5 (heap_path_vptr Hp3 H5).
  destruct H5; destruct H.
  subst l.
  rewrite (mod8_plus4 H6) in H7.
  discriminate.
  apply heap_path_elim with p3; firstorder.

  assert (e : z <> l).
  intro F; subst; apply Nin'.
  apply in_or_app.
  firstorder.
  rewrite Hz in Happ'.
  rewrite app_x_nil_app in Happ'.
  destruct (heap_path_break Happ' Hp1) as [_ Hzl].
  destruct Hzl as [Hv [Ht _]].
  destruct Ht; [left | right];
    rewrite heap_remove_get_ne; trivial.
  intro F.
  destruct Hv; destruct H.
  subst l.
  rewrite (mod8_plus4 H7) in H8.
  discriminate.

  (* case 1.2.2.2: ~ l \in p1 *)
  destruct (list_last_dec p1) as [| [p3 [z Hz]]].

  (* case 1.2.2.2.1:  p = nil :: l :: p2 *)
  subst p1; simpl in Hx.
  rewrite Happ in Hx.
  injection Hx; intros; subst x; constructor; trivial.

  (* case 1.2.2.2.2:  p = (p3 ++ z :: nil) ++ l :: p2 *)
  apply rch_next2 with z; trivial.

  destruct (heap_path_eqv_heap (h := h) (h' := heap_remove h l) 
    (p := p1) (x:= x) (y:=z)); trivial.
  subst; rewrite head_app_not_nil in Hx; trivial.
  destruct p3; intro F; discriminate.
  subst p1; rewrite last_app; reflexivity.
  intros.
  assert (e : l0 <> l).
  intro F; subst; tauto.
  repeat (rewrite heap_remove_get_ne; trivial).
  firstorder.
  intro F.
  hypreplace H5 (heap_path_vptr H2 H5).
  destruct H5; destruct H.
  subst l.
  rewrite (mod8_plus4 H6) in H7.
  discriminate.
  apply heap_path_elim with p1; firstorder.

  assert (e : z <> l).
  intro F; subst; apply H4.
  apply in_or_app.
  firstorder.
  rewrite Hz in Happ.
  rewrite app_x_nil_app in Happ.
  destruct (heap_path_break Happ Hp) as [_ Hzl].
  destruct Hzl as [Hv [Ht _]].
  destruct Ht; [left | right];
    rewrite heap_remove_get_ne; trivial.
  intro F.
  destruct Hv; destruct H.
  subst l.
  rewrite (mod8_plus4 H7) in H8.
  discriminate.

  (* case 1.2.2: h(l + 4) = n, goal 2 *)
  apply rch_next with n; trivial.
  right.
  rewrite heap_remove_get_ne; [trivial | omega].

  assert (Hp'' : heap_path h (n :: p2)).
  destruct Hp'; firstorder.
  destruct (heap_path_eqv_heap (h := h) (h' := heap_remove h l) 
    (p := (n :: p2)) (x:= n) (y:=y)); trivial.
  subst; rewrite last_app in Hy; trivial.
  intros.
  assert (e : l0 <> l).
  intro F; subst; tauto.
  repeat (rewrite heap_remove_get_ne; trivial).
  firstorder.
  intro F.
  hypreplace H4 (heap_path_vptr Hp'' H4).
  destruct H4; destruct H.
  subst l.
  rewrite (mod8_plus4 H5) in H6.
  discriminate.
  apply heap_path_elim with (n::p2); firstorder.
  apply (reach_vptry H1).

  (* case 2: ~ l \in p *)
  destruct (heap_path_eqv_heap (h := h) (h' := heap_remove h l) 
    (p := p) (x:= x) (y:=y)); trivial.
  intros.
  assert (e : l0 <> l).
  intro F; subst; tauto.
  repeat (rewrite heap_remove_get_ne; trivial).
  firstorder.
  intro F.
  hypreplace H2 (heap_path_vptr Hp H2).
  destruct H2; destruct H.
  subst l.
  rewrite (mod8_plus4 H3) in H4.
  discriminate.
  apply heap_path_elim with p; firstorder.
Qed.
  
Lemma reach_heap_remove4_reach : forall h l x y,
  vptr l ->
  heap_get h (l + WSIZE) = Some x ->
  reach h x y -> reach (heap_remove h (l + WSIZE)) x y.
Proof.
  intros.
  destruct (heap_path_intro H1) as [p [Hp [Hx Hy]]].
  destruct (natlist_in_dec p l) as [IN | Nin].

  (* case 1: l \in p *)
  destruct (natlist_break IN) as [p1 [p2 [Happ Nin]]].
  destruct (heap_path_break Happ Hp). 

  destruct p2.

  (* case 1.1: p = p1 ++ l :: nil, l = y *)
  subst.
  assert (E := Hy).
  rewrite last_app in E; injection E; intros; subst.
  clear H2 H3 Nin E.
  destruct (natlist_break2 IN) as [p2 [p3 [Happ NIin]]].
  destruct (heap_path_break Happ Hp).
  destruct (list_last_dec p2) as [| [pt [t Ht]]]; subst.
  rewrite Happ in Hx; injection Hx; intros; subst.
  constructor; trivial.
  apply rch_next2 with t; trivial.
  destruct (heap_path_eqv_heap (h := h) (h' := heap_remove h (y + 4)) 
    (p := (pt ++ t :: nil)) (x:= x) (y:=t)); trivial.
  rewrite Happ in Hx; rewrite head_app_not_nil in Hx; trivial.
  destruct pt; intro F; discriminate.
  rewrite last_app; reflexivity.
  intros.
  assert (e : l <> y ).
  intro F; subst; tauto.
  repeat (rewrite heap_remove_get_ne; trivial).
  firstorder.
  omega.
  intro F.
  hypreplace H4 (heap_path_vptr H2 H4).
  destruct H4; destruct H.
  subst.
  rewrite (mod8_plus4 H6) in H5.
  discriminate.
  apply heap_path_elim with (pt ++ t :: nil); firstorder.

  assert (e : t <> y).
  intro F; subst; apply NIin.
  apply in_or_app.
  firstorder.
  rewrite app_x_nil_app in Happ.
  destruct (heap_path_break Happ Hp).
  destruct H5 as [Hv [Ht _]].
  destruct Ht; [left | right];
    rewrite heap_remove_get_ne; trivial.
  intro F.
  destruct Hv; destruct H.
  subst.
  rewrite (mod8_plus4 H8) in H7.
  discriminate.
  omega.

  (* case 1.2: p = p1 ++ l :: n :: p2 *)
  assert (Hp' := H3).
  destruct H3 as [Hl [Hv _]].
  destruct Hv.

  (* case 1.2.1: h(l) = n *)
  apply reach_trans with l.
  destruct (natlist_in_dec p1 l).

  (* case 1.2.1.1: l \in p1 *)
  destruct (natlist_break2 H4) as [p3 [p4 [Happ' Nin']]].
  destruct (heap_path_break Happ Hp) as [Hp1 _].
  
  destruct (list_last_dec p3) as [| [p5 [z Hz]]].

  (* case 1.2.1.1.1:  p1 = nil :: l :: p4 *)
  subst p3; simpl in Hx.
  rewrite Happ in Hx; rewrite Happ' in Hx.
  injection Hx; intros; subst x; constructor; trivial.

  (* case 1.2.1.1.2:  p1 = (p5 ++ z :: nil) ++ l :: p4 *)
  apply rch_next2 with z; trivial.

  destruct (heap_path_break Happ' Hp1) as [Hp3 _]; trivial.
  destruct (heap_path_eqv_heap (h := h) (h' := heap_remove h (l + 4)) 
    (p := p3) (x:= x) (y:=z)); trivial.
  subst; rewrite head_app_not_nil in Hx; trivial.
  rewrite head_app_not_nil in Hx; trivial.
  destruct p5; intro F; discriminate.
  destruct p5; intro F; discriminate.
  subst p3; rewrite last_app; reflexivity.
  intros.
  assert (e : l0 <> l).
  intro F; subst; tauto.
  repeat (rewrite heap_remove_get_ne; trivial).
  firstorder.
  omega.
  intro F.
  hypreplace H5 (heap_path_vptr Hp3 H5).
  destruct H5; destruct H.
  subst l0.
  rewrite (mod8_plus4 H7) in H6.
  discriminate.
  apply heap_path_elim with p3; firstorder.

  assert (e : z <> l).
  intro F; subst; apply Nin'.
  apply in_or_app.
  firstorder.
  rewrite Hz in Happ'.
  rewrite app_x_nil_app in Happ'.
  destruct (heap_path_break Happ' Hp1) as [_ Hzl].
  destruct Hzl as [Hv [Ht _]].
  destruct Ht; [left | right];
    rewrite heap_remove_get_ne; trivial.
  intro F.
  destruct Hv; destruct H.
  subst z.
  rewrite (mod8_plus4 H8) in H7.
  discriminate.
  omega.

  (* case 1.2.1.2: ~ l \in p1 *)
  destruct (list_last_dec p1) as [| [p3 [z Hz]]].

  (* case 1.2.1.2.1:  p = nil :: l :: p2 *)
  subst p1; simpl in Hx.
  rewrite Happ in Hx.
  injection Hx; intros; subst x; constructor; trivial.

  (* case 1.2.1.2.2:  p = (p3 ++ z :: nil) ++ l :: p2 *)
  apply rch_next2 with z; trivial.

  destruct (heap_path_eqv_heap (h := h) (h' := heap_remove h (l + 4)) 
    (p := p1) (x:= x) (y:=z)); trivial.
  subst; rewrite head_app_not_nil in Hx; trivial.
  destruct p3; intro F; discriminate.
  subst p1; rewrite last_app; reflexivity.
  intros.
  assert (e : l0 <> l).
  intro F; subst; tauto.
  repeat (rewrite heap_remove_get_ne; trivial).
  firstorder.
  omega.
  intro F.
  hypreplace H5 (heap_path_vptr H2 H5).
  destruct H5; destruct H.
  subst l0.
  rewrite (mod8_plus4 H7) in H6.
  discriminate.
  apply heap_path_elim with p1; firstorder.

  assert (e : z <> l).
  intro F; subst; apply H4.
  apply in_or_app.
  firstorder.
  rewrite Hz in Happ.
  rewrite app_x_nil_app in Happ.
  destruct (heap_path_break Happ Hp) as [_ Hzl].
  destruct Hzl as [Hv [Ht _]].
  destruct Ht; [left | right];
    rewrite heap_remove_get_ne; trivial.
  intro F.
  destruct Hv; destruct H.
  subst z.
  rewrite (mod8_plus4 H8) in H7.
  discriminate.
  omega.

  (* case 1.2.1: h(l + 4) = n, goal 2 *)
  apply rch_next with n; trivial.
  left.
  rewrite heap_remove_get_ne; [trivial | omega].

  assert (Hp'' : heap_path h (n :: p2)).
  destruct Hp'; firstorder.
  destruct (heap_path_eqv_heap (h := h) (h' := heap_remove h (l + 4)) 
    (p := (n :: p2)) (x:= n) (y:=y)); trivial.
  subst; rewrite last_app in Hy; trivial.
  intros.
  assert (e : l0 <> l).
  intro F; subst; tauto.
  repeat (rewrite heap_remove_get_ne; trivial).
  firstorder.
  omega.
  intro F.
  hypreplace H4 (heap_path_vptr Hp'' H4).
  destruct H4; destruct H.
  subst l0.
  rewrite (mod8_plus4 H6) in H5.
  discriminate.
  apply heap_path_elim with (n::p2); firstorder.
  apply (reach_vptry H1).

  (* case 1.2.1: n = x *)
  rewrite H3 in H0; injection H0; intros; subst.
  assert (Happ := app_x_nil_app p1 (x ::p2) l). 
  destruct (heap_path_break (sym_eq Happ) Hp).
  destruct (heap_path_eqv_heap (h := h) (h' := heap_remove h (l + 4)) 
    (p := (x :: p2)) (x:= x) (y:=y)); trivial.
  rewrite <- Happ in Hy; rewrite last_app in Hy; trivial.
  intros.
  assert (e : l0 <> l).
  intro F; subst; tauto.
  repeat (rewrite heap_remove_get_ne; trivial).
  firstorder.
  omega.
  intro F.
  hypreplace H6 (heap_path_vptr H5 H6).
  destruct H6; destruct H.
  subst.
  rewrite (mod8_plus4 H8) in H7.
  discriminate.
  apply heap_path_elim with (x :: p2); firstorder.

  (* case 2: ~ l \in p *)
  destruct (heap_path_eqv_heap (h := h) (h' := heap_remove h (l + 4)) 
    (p := p) (x:= x) (y:=y)); trivial.
  intros.
  assert (e : l0 <> l).
  intro F; subst; tauto.
  repeat (rewrite heap_remove_get_ne; trivial).
  firstorder.
  omega.
  intro F.
  hypreplace H2 (heap_path_vptr Hp H2).
  destruct H2; destruct H.
  subst l0.
  rewrite (mod8_plus4 H4) in H3.
  discriminate.
  apply heap_path_elim with p; firstorder.
Qed.


(** ******************************************************************* *)
(** computational reachability *)
(** ******************************************************************* *)
Fixpoint reachb_aux (n : nat) (hp : heap) (x y : nat) {struct n} : bool :=
  if beq_nat x y then vptrb x else
    match n with
      | S n' =>
	vptrb x && vptrb y &&
	(match (heap_get hp x) with
	   | Some z => reachb_aux n' (heap_remove hp x) z y
	   | _ => false
	 end ||
	match (heap_get hp (x + WSIZE)) with
	  | Some z => reachb_aux n' (heap_remove hp (x + WSIZE)) z y
	  | _ => false
	end)
      | _ => false
    end.

Notation reachb := (fun h x y => reachb_aux (heap_size h) h x y).

(** lemmas: reachb *)
Lemma reachb_elim : forall h x y, reachb h x y = true -> reach h x y.
Proof.
  intros h x y.
  assert (exists n, heap_size h = n).
  exists (heap_size h); reflexivity.
  destruct H as [n Hn].
  generalize n h x y Hn; clear n h x y Hn.

  induction n; intros h x y Hn; rewrite Hn.
  simpl; simplif; intro.
  constructor; auto with vptr.
  discriminate.
  simpl; simplif.
  constructor; auto with vptr.
  destruct (heap_get_dec2 h x) as [Hx | [vx Hx]];
    destruct (heap_get_dec2 h (x + 4)) as [Hx4 | [vx4 Hx4]];
      try rewrite Hx; rewrite Hx4; simpl; intros; 
	elimbool; try discriminate.

  apply rch_next with vx4; auto with vptr.
  apply heap_remove_reach with (x + 4); trivial.
  hypreplace Hn (heap_remove_size_dcs (heap_get_some_heap_in Hx4) Hn).
  subst n; firstorder.

  apply rch_next with vx; auto with vptr.
  apply heap_remove_reach with x; trivial.
  hypreplace Hn (heap_remove_size_dcs (heap_get_some_heap_in Hx) Hn).
  subst n; firstorder.
  firstorder.

  apply rch_next with vx; auto with vptr.
  apply heap_remove_reach with x; trivial.
  hypreplace Hn (heap_remove_size_dcs (heap_get_some_heap_in Hx) Hn).
  subst n; firstorder.
  firstorder.
	
  apply rch_next with vx4; auto with vptr.
  apply heap_remove_reach with (x + 4); trivial.
  hypreplace Hn (heap_remove_size_dcs (heap_get_some_heap_in Hx4) Hn).
  subst n; firstorder.
Qed.


Lemma reachb_intro : forall h x y, reach h x y -> reachb h x y = true.
Proof.
  intros h x y.
  assert (exists n, heap_size h = n).
  exists (heap_size h); reflexivity.
  destruct H as [n Hn].
  generalize n h x y Hn; clear n h x y Hn.

  induction n; intros h x y Hn Hrch; rewrite Hn.

  (* case 1: heap_size h = 0 *)
  induction Hrch.
  simpl; simplbeq; auto with vptr.
  hypreplace Hn (heap_size_0_empty_heap Hn).
  subst; simpl.
  repeat (rewrite heap_get_empty_heap_None in H).
  destruct H; discriminate.


  (* case 2: heap_size h = S n *)
  compare x y; intro e; subst.
  inversion_clear Hrch;
    simpl; simplbeq; auto with vptr.

  inversion Hrch; subst.
  destruct e; reflexivity.
  
  simpl; simplbeq; splitb; auto with vptr.
  destruct H; rewrite H; [left | right].
  
  hypreplace H0 (reach_heap_remove0_reach H1 H H0).
  hypreplace Hn (heap_remove_size_dcs (heap_get_some_heap_in H) Hn).
  rewrite <- Hn; apply (IHn _ _ _ Hn H0).

  hypreplace H0 (reach_heap_remove4_reach H1 H H0).
  hypreplace Hn (heap_remove_size_dcs (heap_get_some_heap_in H) Hn).
  rewrite <- Hn; apply (IHn _ _ _ Hn H0).
Qed.

Lemma reachb_dec : forall h x y,
  reachb h x y = true \/ reachb h x y = false.
Proof.
  intros.
  apply bool_dec.
Qed.

Lemma reachb_intro_false : forall h x y,
  ~ reach h x y -> reachb h x y = false.
Proof.
  intros.
  destruct (reachb_dec h x y); trivial.
  destruct H; apply (reachb_elim H0).
Qed.

Lemma reachb_elim_false : forall h x y,
  reachb h x y = false -> ~ reach h x y.
Proof.
  red; intros.
  hypreplace H0 (reachb_intro H0).
  destruct (reachb h x y); discriminate.
Qed.

Lemma reach_dec : forall h x y,
  reach h x y \/ ~ reach h x y.
Proof.
  intros.
  destruct (reachb_dec h x y).
  left; apply (reachb_elim H).
  right; apply (reachb_elim_false H).
Qed.


(** ******************************************************************* *)
(** reachability from a set of pointers *)
(** ******************************************************************* *)
Notation reachs := (fun hp s x => 
  (exists a, In a s /\ reach hp a x)).

Fixpoint reachsb (hp : heap) (s : list nat) (x : nat) {struct s} : bool :=
  match s with
    | a :: s' => (reachb hp a x) || (reachsb hp s' x)
    | _ => false
  end.

(* lemmas: reachsb *)
Lemma reachsb_intro : forall s hp x,
  reachs hp s x -> reachsb hp s x = true.
Proof.
  induction s; simpl; intros hp x [b [Hin Hrch]].
  tauto.
  splitb.
  destruct Hin; [left | right].
  subst; apply reachb_intro; trivial.
  apply IHs.
  exists b; firstorder.
Qed.

Lemma reachsb_elim : forall s hp x,
  reachsb hp s x = true -> reachs hp s x.
Proof.
  induction s; simpl; intros.
  discriminate.
  elimbool.
  exists a; split.
  firstorder.
  apply reachb_elim; trivial.
  destruct (IHs _ _ B0).
  exists x0; firstorder.
Qed.

Lemma reachsb_dec : forall hp s x,
  reachsb hp s x = true \/ reachsb hp s x = false.
Proof.
  intros.
  apply bool_dec.
Qed.

Lemma reachsb_elim_false : forall s hp x,
  reachsb hp s x = false -> ~ reachs hp s x.
Proof.
  red; intros.
  hypreplace H0 (reachsb_intro H0).
  rewrite H in H0; discriminate.
Qed.

Lemma reachsb_intro_false : forall s hp x,
  ~ reachs hp s x -> reachsb hp s x = false.
Proof.
  intros.
  destruct (reachsb_dec hp s x); trivial.
  hypreplace H0 (reachsb_elim H0).
  tauto.
Qed.
