(** SCAP on LOCAP. *)

(** Chunxiao Lin @ 2006 *)

Set Implicit Arguments.
Unset Strict Implicit.

Require Export locap.


(** ******************************************************************* *)
(** typing rules *)
(** ******************************************************************* *)

(** well-formed instruction sequence rule *)
Fixpoint scap_iseq_ok (P : chspec) (p : spred) (g : guar) (is : iseq) {struct is} : Prop :=
  match is with
    | jr r31 => forall s, p s -> g s s
    | j target => exists p', exists g',
      (chspec_in P target (isp2 p' g')) /\
      (forall s, p s -> p' s /\ forall s', g' s s' -> g s s')
    | jal fret target => exists p', exists g', exists p'', exists g'',
      (chspec_in P target (isp2 p' g')) /\
      (chspec_in P fret (isp2  p'' g'')) /\
      (forall s, p s -> p' (rset s r31 fret) /\ 
	forall s', g' (rset s r31 fret) s' -> p'' s' /\
	  forall s'', g'' s' s'' -> g s s'') /\
      (forall s s', g' s s' -> rget s r31 = rget s' r31)
    | br bt rs rt offset is' => exists p'', exists g'',
      scap_iseq_ok P p'' g'' is' /\
      forall s, p s ->
	if btest bt (rget s rs) (rget s rt) 
	  then exists p', exists g', 
	    (chspec_in P offset (isp2 p' g')) /\
	    p' s /\ (forall s', g' s s' -> g s s')
	  else p'' s /\ forall s', g'' s s' -> g s s'
    | seq c is' => exists p', exists g' : guar,
      (forall s, p s -> opt_pred (cstep c s) (fun s' =>
	p' s' /\ forall s'', g' s' s'' -> g s s'')) /\
      scap_iseq_ok P p' g' is'
    | jr _ => False
  end.


(** ******************************************************************* *)
(** weakening lemma *)
(** ******************************************************************* *)
Lemma scap_iseq_ok_weakening : forall is P (p p' : spred) (g g' : guar),
  (forall s, p s -> p' s /\ forall s', g' s s' -> g s s') ->
    scap_iseq_ok P p' g' is -> scap_iseq_ok P p g is.
Proof.
  induction is; intros.
  destruct H0 as [p0 [g0 [Hc His']]].
  fold scap_iseq_ok in His'.
  exists p0; exists g0.
  split; trivial.
  intros.
  destruct (H _ H0).
  hypreplace Hc (Hc _ H1).
  destruct (cstep c s) as [s' | ]; intuition.
  
  destruct H0 as [p1 [g1 Hok]].
  exists p1; exists g1.
  fold scap_iseq_ok in Hok.
  intuition.
  destruct (H _ H2).
  hypreplace H1 (H1 _ H3).
  destruct (btest b (rget s r) (rget s r0));
    intuition.
  destruct H1 as [p0 [g0 [Hpn [Hp' Hg']]]].
  exists p0; exists g0; intuition.
  
  destruct H0 as [p0 [g0 [Hpsi His]]].
  exists p0; exists g0.
  split.
  exact Hpsi.
  intros.
  destruct (H s H0).
  destruct (His s H1).
  split.
  exact H3.
  intuition.
  
  destruct H0 as [p0 [g0 [p1 [g1 [Hpsi1 [Hpsi2 [His1 His2]]]]]]].
  exists p0; exists g0; exists p1; exists g1.
  repeat (split; trivial).
  destruct (H s H0).
  destruct (His1 s H1).
  exact H3.
  destruct (H s H0).
  destruct (His1 s H2).
  destruct (H5 s' H1).
  exact H6.
  destruct (H s H0).
  destruct (His1 s H2).
  destruct (H5 s' H1).
  intuition.

  destruct r; try (elimtype False; apply H0).
  simpl in *.
  intros.
  destruct (H _ H1).
  intuition.
Qed.


(** ******************************************************************* *)
(** soundness *)
(** ******************************************************************* *)
Theorem scap_cap_sound_iseq : forall is P p g,
  scap_iseq_ok P p g is -> 
  iseq_ok (lift (interp (isp2 p g)) P) is.
Proof.
  induction is; intros P p g Hok.

  destruct Hok as [p' [g' [Hs Hok']]].
  fold scap_iseq_ok in Hok'.
  exists (lift (interp (isp2 p' g')) P).
  split; [idtac | apply IHis; trivial].
  unfold comm_ok.
  intros Pglob s Hp.
  destruct Hp as [Hpsi [Hp [n Hwf]]].
  hypreplace Hs (Hs _ Hp).
  destruct (cstep c s) as [s' |]; 
    [destruct Hs as [Hp' Hs''] | elimtype False; exact Hs].
  repeat (split; trivial).
  exists n.
  apply (WFST_weakening Hs'' Hwf).

  destruct Hok as [p'' [g'' [His' Hok]]].
  fold scap_iseq_ok in * .
  exists (lift (interp (isp2 p'' g'')) P).
  split.
  apply IHis.
  apply His'.
  intros Pglob s [Hpsi [Hp [nn Hwf]]].
  hypreplace Hok (Hok _ Hp).
  destruct (btest b (rget s r) (rget s r0)).
  destruct Hok as [p' [g' [Hpn [Hp' Hg']]]].
  exists (isp2 p' g').
  split.
  apply (Hpsi _ _ Hpn).
  split; trivial.
  exists nn.
  apply (WFST_weakening Hg' Hwf).
  destruct Hok.
  repeat (split; trivial).
  exists nn.
  apply (WFST_weakening H0 Hwf).

  destruct Hok as [p' [g' [Hn Hok]]].
  intros Pglob s [Hpsi [Hp [nn Hwf]]].
  exists (isp2 p' g').
  split.
  apply (Hpsi _ _ Hn).
  destruct (Hok _ Hp).
  split; trivial.
  exists nn; apply (WFST_weakening H0 Hwf).

  destruct Hok as [p' [g' [p'' [g'' [Hn [Hn' [Hok Hra]]]]]]].
  intros Pglob s [Hpsi [Hp [nn Hwf]]].
  exists (isp2 p' g').
  split.
  apply (Hpsi _ _ Hn).
  destruct (Hok _ Hp) as [Hsr Hk].
  split; trivial.
  exists (S nn).
  unfold WFST; fold WFST. 
  intros s' Hg'.
  rewrite <- (Hra (rset s r31 n) s').
  simplst.
  exists p''; exists g''.
  split.
  apply (Hpsi _ _ Hn').
  destruct (Hk _ Hg').
  split; trivial.
  apply (WFST_weakening H0 Hwf).
  apply Hg'.

  destruct r; try (elimtype False; apply Hok).
  intros Pglob s [Hpsi [Hp [nn Hwf]]].
  assert (Hok' := Hok _ Hp).
  destruct nn.
  simpl in Hwf.
  destruct (Hwf _ Hok') as [G [HG1 HG2]].
  exists (isp1 G); split; trivial.
  simpl in Hwf.
  destruct (Hwf _ Hok') as [p' [g' [X1 [X2 X3]]]].
  exists (isp2 p' g').
  repeat (split; trivial).
  exists nn; apply X3.
Qed.


