(** Separation logic *)

(** Chunxiao Lin @ 2006 *)

Require Export heap2.
Require Export natset2.
Require Import basicutil.

Set Implicit Arguments.
Unset Strict Implicit.


(** ******************************************************************* *)
(** heap predicate *)
(** ******************************************************************* *)
Notation hpred := (heap -> Prop).

Definition hpred_true (h:heap) (A:hpred) := A h.

Notation "h ||- A" := (hpred_true h A) (at level 80).


(** ******************************************************************* *)
(** heap eq *)
(** ******************************************************************* *)
Definition hp_eq (h:heap) : hpred:= fun h' => h' = h.


(** ******************************************************************* *)
(** separation conjunction *)
(** ******************************************************************* *)
Definition sconj (A B : hpred) : hpred :=
  fun h => exists ha, exists hb,
    heap_join ha hb h /\
    ha ||- A /\ hb ||- B.

Infix "**" := sconj (at level 60, right associativity).


(** ******************************************************************* *)
(** alternative conjuction *)
(** ******************************************************************* *)
Definition aconj (A B:hpred) : hpred :=
  fun h => A h /\ B h.

Infix "&" := aconj (at level 60, right associativity).


(** ******************************************************************* *)
(** emp *)
(** ******************************************************************* *)
Definition unit : hpred := heap_eqv empty_heap.


(** ******************************************************************* *)
(** true *)
(** ******************************************************************* *)
Definition top : hpred := fun h => True.


(** ******************************************************************* *)
(** \exists x, f x *)
(** ******************************************************************* *)
Definition lex (A:Set) (f:A -> hpred) : hpred :=
  fun h => exists x, f x h.

Notation "'lexists' x , A" := (lex (fun x => A))
  (at level 200, x ident).


(** ******************************************************************* *)
(** !(P) *)
(** ******************************************************************* *)
Definition bang (A:Prop) : hpred :=
  fun h => unit h /\ A.

Notation "! A" := (bang A) (at level 55).


(** ******************************************************************* *)
(** singleton heap predicate *)
(** ******************************************************************* *)
Definition cont (l:nat) (x:nat) : hpred :=
  fun h => ptr l /\ h = (singleton_heap l x).

Infix "|->" := cont (at level 59, right associativity).


(** ******************************************************************* *)
(** iterated separation conjunction *)
(** ******************************************************************* *)
Fixpoint itr (p:nat -> hpred) (l : list nat) {struct l} : hpred :=
  match l with
    | nil => unit
    | cons x l' => (p x) ** (itr p l')
  end.

Definition fas (p:nat -> hpred) (s: set) : hpred :=
  itr p (sd_of s).

Notation "'foralls' x 'in' s , A" := (fas (fun x => A) s) 
  (at level 200, x ident).


(** ******************************************************************* *)
(** l |-> x1, ..., xn *)
(** ******************************************************************* *)
Fixpoint contL (l:nat) (xx:list nat) {struct xx} : hpred :=
  match xx with
    | nil => unit
    | x::nil => l |-> x
    | x::xx' => l |-> x ** contL (4 + l) xx'
  end.

Infix "|->," := contL (at level 59, right associativity).

Definition pairCont x v0 v1 := x |->, (v0::v1::nil).

Infix "|->>" := pairCont (at level 59, right associativity).

(* l |-> _ *)
Notation "l |->_" := (lexists xyzyzy, l |-> xyzyzy) (at level 59).
(* l |-> _, _ *)
Notation "l |->__" := (((l) |->_) ** (S l |->_)) (at level 59).
(* l |-> _, _, _ *)
Notation "l |->___" := (((l) |->_) ** (S l |->__)) (at level 59).
(* l |-> _, _, _, _ *)
Notation "l |->____" := (((l) |->_) ** (S l |->___)) (at level 59).

Notation "A ==> B" :=
  (forall h, h ||- A -> h ||- B) (at level 13).


(** ******************************************************************* *)
(** lemmas *)
(** ******************************************************************* *)
Section SConj.

  Lemma sconj_intro : forall ha hb h A B,
    ha ||- A -> hb ||- B -> heap_join ha hb h ->
      h ||- A ** B.
  Proof.
    intros.
    exists ha; exists hb; auto.
  Qed.

  (* elim is unfold *)

  Lemma sconj_sym : forall h A B,
    h ||- A ** B -> h ||- B ** A. 
  Proof.  
    intros h A B [ha [hb [J [Ha Hb]]]].
    apply (sconj_intro Hb Ha (heap_join_sym J)).
  Qed.

  Lemma sconj_mono_left : forall h A B,
    h ||- (A ** B) -> 
      forall A', (A ==> A') -> 
      h ||- (A' ** B).
  Proof.
    destruct 1 as [ha [hb [J [HA HB]]]].
    exists ha; exists hb.
    intuition.
  Qed.

  Lemma sconj_mono_right : forall h A B,
    h ||- (A ** B) -> 
      forall B', (B ==> B') -> 
      h ||- (A ** B').
  Proof.
    destruct 1 as [ha [hb [J [HA HB]]]].
    exists ha; exists hb.
    intuition.
  Qed.

  Lemma sconj_mono : forall h A A' B B',
    h ||- A ** B -> 
      (A ==> A') -> (B ==> B') -> 
      h ||- A' ** B'.
  Proof.
    intros.
    assert (h ||- (A' ** B)).
    eapply sconj_mono_left; [eauto | assumption].
    eapply sconj_mono_right; [eauto | assumption].
  Qed.

  Lemma sconj_assoc_left : forall A B C h,
    h ||- (A ** B) ** C -> h ||- A ** (B ** C).
  Proof.
    intros.
    destruct H as [hab [hc [Jabc [HAB HC]]]].
    destruct HAB as [ha [hb [Jab [HA HB]]]].
    destruct (heap_join_assoc Jab Jabc) as [hbc [Jabc' Jbc]].
    eapply sconj_intro; eauto.
    eapply sconj_intro; eauto.
  Qed.

  Lemma sconj_assoc_right : forall A B C h,
    h ||- A ** (B ** C) -> h ||- (A ** B) ** C.
  Proof.
    intros.
    apply sconj_sym.
    apply sconj_assoc_left.
    exact (sconj_sym (sconj_assoc_left (sconj_sym H))).
  Qed.

  Lemma sconj_insert : forall h h' B A,
    h ||- hp_eq h' ** B -> h' ||- A -> h ||- A ** B.
  Proof.
    intros h ha' B H A HA.
    apply (sconj_mono_left A); intros.
    unfold hpred_true in H0.
    rewrite H0; trivial.
  Qed.


End SConj.


Section Unit.

  Lemma unit_intro : empty_heap ||- unit.
  Proof.
    unfold hpred_true.
    unfold unit. 
    apply heap_eqv_refl.
  Qed.

  Lemma unit_intro_left : forall h P,
    h ||- P -> h ||- unit ** P.
  Proof.
    intros.
    eapply sconj_intro.
    2: apply H.
    apply unit_intro.
    intro addr. 
    destruct (heap_get h addr); simpl; trivial.
  Qed.

  Lemma unit_intro_right : forall h P,
    h ||- P -> h ||- P ** unit.
  Proof.
    intros.
    apply sconj_sym.
    apply (unit_intro_left H).
  Qed.

  Lemma unit_elim : forall h,
    h ||- unit -> h = empty_heap.
  Proof.
    intros.
    symmetry.
    apply (heap_ext H).
  Qed.

  Lemma unit_elim_left : forall h P,
    h ||- unit ** P -> h ||- P.
  Proof.
    intros.
    destruct H as [h1 [h2 [HJ [H1 H2]]]].
    hypreplace HJ (heap_join_elim3 (heap_eqv_sym H1) HJ).
    rewrite (heap_ext HJ) in H2.
    apply H2.
  Qed.

  Lemma unit_elim_right : forall h P,
    h ||- P ** unit -> h ||- P.
  Proof.
    intros.
    apply (unit_elim_left (sconj_sym H)).
  Qed.

End Unit.

Section Top.

  Lemma top_intro : forall h, h ||- top.
  Proof.
    unfold hpred_true.
    unfold top. 
    tauto.
  Qed.

  Lemma top_intro_left : forall h P,
    h ||- P -> h ||- top ** P.
  Proof.
    intros.
    assert (heap_join empty_heap h h).
    intro addr. 
    destruct (heap_get h addr); simpl; trivial.
    apply (sconj_intro (top_intro empty_heap) H H0).
  Qed.

  Lemma top_intro_right : forall h P,
    h ||- P -> h ||- P ** top.
  Proof.
    intros.
    apply sconj_sym.
    apply (top_intro_left H).
  Qed.

  Lemma top_elim_left : forall h P,
    h ||- top ** P -> exists h', h ||- hp_eq h' ** P.
  Proof.
    intros.
    destruct H as [ha [hb [J [Ha Hb]]]].
    exists ha.
    apply sconj_intro with ha hb; auto.
    unfold hp_eq.
    unfold hpred_true; reflexivity.
  Qed.

  Lemma top_elim_right : forall h P,
    h ||- P ** top -> exists h', h ||- P ** hp_eq h'.
  Proof.
    intros.
    destruct H as [ha [hb [J [Ha Hb]]]].
    exists hb.
    apply sconj_intro with ha hb; auto.
    unfold hp_eq.
    unfold hpred_true; reflexivity.
  Qed.

End Top.

Section Bang.

  Lemma bang_intro : forall h (A:Prop),
    h ||- unit -> A ->
      h ||- ! A.
  Proof.
    unfold bang.
    intros h A Hp Aok.
    constructor; assumption.
  Qed.
  
  Lemma bang_elim1 : forall h A,
    h ||- ! A -> A.
  Proof.
    unfold bang.
    unfold hpred_true.
    intuition.
  Qed.

  Lemma bang_elim2 : forall h A,
    h ||- ! A -> h ||- unit.
  Proof.
    unfold bang; unfold hpred_true.
    intuition.
  Qed.

  Lemma bang_elim : forall h A,
    h ||- ! A ->
      h ||- unit /\ A.
  Proof.
    intros h A Hp.
    constructor.
    exact (bang_elim2 Hp).
    exact (bang_elim1 Hp).
  Qed.

  Lemma bang_elim_left : forall h A B,
    h ||- ! A ** B ->
      A /\ h ||- B.
  Proof.
    intros h A B H.
    assert (A /\ h ||- unit ** B).
    destruct H as [ha [hb [J [HA HB]]]].
    constructor.
    apply (bang_elim1 HA).
    assert (HA' := bang_elim2 HA).
    eapply sconj_intro; eauto.
    intuition.
    exact (unit_elim_left H2).
  Qed.

  Lemma bang_elim_left1 : forall h A B,
    h ||- ! A ** B ->
      A.
  Proof.
    intros h A B H.
    destruct (bang_elim_left H).
    assumption.
  Qed.

  Lemma bang_elim_left2 : forall h A B,
    h ||- ! A ** B ->
      h ||- B.
  Proof.
    intros h A B H.
    destruct (bang_elim_left H).
    assumption.
  Qed.

  Lemma bang_elim_right : forall h A B,
    h ||- B ** ! A ->
      A /\ h ||- B.
  Proof.
    intros h A B H.
    hypreplace H (sconj_sym H).
    apply bang_elim_left; auto.
  Qed.

  Lemma bang_intro_left : forall h (A:Prop) B,
    A -> h ||- B ->
      h ||- ! A ** B.
  Proof.
    intros h A B HA HB.
    unfold bang.
    assert (Q := unit_intro_left HB).
    eapply sconj_mono; unfold hpred_true; eauto.
    assumption.
  Qed.

  Lemma bang_intro_right : forall h (A:Prop) B,
    A -> h ||- B ->
      h ||- B ** ! A.
  Proof.
    intros.
    apply sconj_sym.
    apply bang_intro_left; auto.
  Qed.

End Bang.

Section Cont.

  Lemma cont_elim1 : forall h l w,
    h ||- l |-> w -> h = (singleton_heap l w).
  Proof.
    unfold hpred_true; unfold cont; simpl.
    tauto.
  Qed.

  Lemma cont_elim2 : forall h l w,
    h ||- l |-> w -> ptr l.
  Proof.
    unfold hpred_true; unfold cont; simpl.
    tauto.
  Qed.

  Lemma cont_get_eq : forall h l w,
    h ||- l |-> w -> 
      heap_get h l = Some w.
  Proof.
    intros h l w H.
    rewrite (cont_elim1 H).
    unfold singleton_heap. 
    simplhp.
    reflexivity.
    apply (cont_elim2 H).
  Qed.
    
  Lemma sconjDisj : forall x y z h,
    h ||- x |-> y ** x |-> z ->
      False.
  Proof.
    intros.
    destruct H as [h1 [h2 [J [H1 H2]]]].
    
    hypreplace H1 (hd_get_some_hd_in (cont_get_eq H1)).
    hypreplace H2 (hd_get_some_hd_in (cont_get_eq H2)).
    hypreplace J (heap_join_elim1 J).
    elimtype False.
    apply (J x H1 H2).
  Qed.

  Lemma cont_get_eq1 : forall h l w A,
    h ||- l |-> w ** A -> 
      heap_get h l = Some w.
  Proof.
    intros h l w A Q.
    destruct Q as [ha [hb [J [HA HB]]]].
    hypreplace HA (cont_get_eq HA).
    apply (heap_join_sound_left J HA).
  Qed.

  Lemma cont_update : forall a v1 v2 h,
    h ||- a |-> v1 ->
      heap_set h a v2 ||- a |-> v2.
  Proof.
    unfold cont; unfold singleton_heap; 
      unfold hpred_true; simpl.  
    intuition.
    rewrite H1.
    apply heap_ext; intros x.
    compare x a; intro e; subst;
      simplhp; auto.
  Qed.

  Lemma cont_update1 : forall h a v1 A,
    h ||- (a |-> v1 ** A) -> forall v2,
      heap_set h a v2 ||- (a |-> v2 ** A).
  Proof.
    intros h a v1 A [h1 [h2 [J [H1 H2]]]] v2.
    exists (heap_set h1 a v2).
    exists h2.
    split.
    intro x.
    compare a x; intro e;[subst a | idtac]; simplhp.
    hypreplace H1 (hd_get_some_hd_in (cont_get_eq H1)).
    hypreplace J (heap_join_elim1 J  H1).
    destruct (heap_get_dec h2 x).
    rewrite H.
    constructor.
    assert (exists n,hd_get (hd_of h2) x = Some n).
    unfold heap_get in * .
    destruct (hd_get (hd_of h2) x).
    exists n; reflexivity.
    tauto.
    destruct H0.
    hypreplace H0 (hd_get_some_hd_in H0).
    elimtype False; apply (J H0).
    unfold hpred_true in H1; unfold cont in H1; intuition.
    unfold hpred_true in H1; unfold cont in H1; intuition.
    apply (J x).
    split; trivial.
    apply (cont_update v2 H1).
  Qed.

  Lemma join_sing_eq : forall h h1 h2 h3 h4 a v,
    h1 ||- a |-> v -> heap_join h1 h2 h ->
      h3 ||- a |-> v -> heap_join h3 h4 h ->
	h2 = h4.
  Proof.
    intros.
    destruct H; destruct H1.  
    apply heap_ext; intro x.
    compare a x; intro R; subst;
    hypreplace H0 (H0 x); hypreplace H2 (H2 x); generalize H0 H2;
      unfold singleton_heap; unfold empty_heap; simpl;
	simplhp; auto; clear H0 H2;
	  destruct (heap_get h2 x); destruct (heap_get h4 x);
	    destruct (heap_get h x); intros;
	      try (elimtype False; apply H0);
		try (elimtype False; apply H2);
		  try discriminate; try reflexivity; auto. 
    inversion_clear H2; inversion_clear H0.
    reflexivity.
  Qed.

  Lemma cont_cont_eq : forall h x v1 v2,
    h ||- x |-> v1 -> 
      h ||- x |-> v2 -> 
	v1 = v2.
  Proof.
    intros h x v1 v2 H1 H2.
    assert (Some v1 = Some v2).
    rewrite <- (cont_get_eq H1).
    rewrite <- (cont_get_eq H2).
    reflexivity.
    injection H; auto.
  Qed.

  Lemma cont_cont_eq1 : forall h x v1 v2 A B,
    h ||- x |-> v1 ** A -> 
      h ||- x |-> v2 ** B -> 
	v1 = v2.
  Proof.
    intros h x v1 v2 A B H1 H2.
    assert (Some v1 = Some v2).
    rewrite <- (cont_get_eq1 H1).
    rewrite <- (cont_get_eq1 H2).
    reflexivity.
    injection H; auto.
  Qed.

  Lemma cont_cont_false : forall h a x y A,
    h ||-  a |-> x ** a |-> y ** A -> False.
  Proof.
    intros.
    destruct H as [h1 [h2 [J [Hp1 Hp2]]]].
    hypreplace J (J a).
    rewrite (cont_get_eq Hp1) in J.
    rewrite (cont_get_eq1 Hp2) in J.
    destruct (heap_get h a); inversion_clear J.
  Qed.

End Cont.

Section Lex.

  Lemma lex_intro : forall h (A:Set) (f:A->hpred) x,
    h ||- f x ->
      h ||- lex f.
  Proof.
    intros.
    exists x.
    assumption.
  Qed.

  Lemma lex_intro1 : forall h (A:Set) (f:A->hpred),
    (exists x, h ||- f x) ->
    h ||- lex f.
  Proof.
    intros h A f [x Hp].
    apply (lex_intro (A := A) Hp).
  Qed.

  Lemma lex_elim : forall h (A:Set) (f:A->hpred),
    h ||- lex f -> 
      exists x, h ||- f x.
  Proof.
    unfold lex; auto.
  Qed.

  Lemma lex_elim_left  : forall (T:Set) h (A:T -> hpred) B,
    h ||- (lex A) ** B -> h ||- lexists x, (A x) ** B.
  Proof.
    intros T h A B H.
    destruct H as [h1 [h2 [J [HA HB]]]].
    destruct (lex_elim HA) as [x HA']; clear HA.
    eapply lex_intro; eapply sconj_intro; eauto.
  Qed.

  Lemma lex_elim_right : forall (T:Set) h A (B:T -> hpred),
    h ||- A ** (lex B) -> h ||- lexists x, A ** (B x).
  Proof.
    intros T h A B H.
    destruct (lex_elim_left (sconj_sym H)) as [x H'].
    exists x.
    apply sconj_sym; assumption.
  Qed.

  Lemma lex_intro_left : forall (T:Set) h A (B:T -> hpred),
    h ||- (lexists x, (B x) ** A) -> h ||- (lex B) ** A.
  Proof.
    intros T h A B Hp.
    destruct (lex_elim Hp) as [x Hp'].
    eapply sconj_mono_left; eauto.
    intros; eapply lex_intro; eauto.
  Qed.

  Lemma lex_intro_right : forall (T:Set) h A (B:T -> hpred),
    h ||- (lexists x, A ** (B x)) -> h ||- A ** (lex B).
  Proof.
    intros T h A B Hp.
    destruct (lex_elim Hp) as [x Hp'].
    eapply sconj_mono_right; eauto.
    intros; eapply lex_intro; eauto.
  Qed.

End Lex.

Section Fas.

  Lemma fas_intro2 : forall h P,
    h ||- unit ->
      h ||- fas P empty_set.
  Proof.
    unfold fas.
    simpl.
    auto.
  Qed.

  Lemma fas_elim2 : forall h P,
    h ||- fas P empty_set ->
      h ||- unit.
  Proof.
    unfold fas.
    simpl.
    auto.
  Qed.

  Lemma fas_intro1 : forall s a P h,
    h ||- P a ** fas P s -> ~ set_in a s ->
      h ||- fas P (set_add a s).
  Proof.
    intro s; destruct s; 
      induction s as [ | b s]; intros.
    apply H.
    unfold fas in *; unfold set_in in *; 
      simpl in *; intros.
    simplcmp a b.
    apply H.
    hypreplace RE (beq_nat_eq _ _ (sym_eq RE)).
    elimtype False; intuition.
    simpl.
    apply (sconj_mono_right 
      (sconj_assoc_left 
	(sconj_sym H))).
    intros.
    apply IHs; auto.
    destruct (andb_true_eq _ _ (sym_eq e)).
    symmetry; exact H3.
    apply (sconj_sym H1).
  Qed.

  Lemma fas_elim1 : forall s a P h,
    set_in a s -> h ||- fas P s ->
      h ||- P a ** fas P (set_remove a s).
  Proof.
    intro s; destruct s; 
      induction s as [ | b s]; intros.
    elimtype False; apply H.
    unfold fas in *; unfold set_in in *;
      unfold set_remove in *;
	simpl in *; intros.
    simpl in e.
    destruct (andb_true_eq _ _ (sym_eq e)); clear e.
    destruct H.
    subst b; simplbeq.
    rewrite (sd_remove_none (lt_sd_not_in (blt_sd_elim (sym_eq H1)))).
    apply H0.
    compare a b; intro e; [subst b | idtac]; simplbeq.   
    rewrite (sd_remove_none (lt_sd_not_in (blt_sd_elim (sym_eq H1)))).
    apply H0.
    simpl.
    apply sconj_sym. 
    apply sconj_assoc_right.
    apply (sconj_mono_right H0).
    intros.
    apply sconj_sym.
    apply IHs; auto.
  Qed.

  Lemma fas_elim : forall s a P h A,
    set_in a s -> h ||- fas P s ** A ->
      h ||- P a ** fas P (set_remove a s) ** A.
  Proof.
    intros.
    apply sconj_assoc_left; 
      apply (sconj_mono_left H0); intros.
    apply fas_elim1; trivial.
  Qed.

  Lemma fas_intro : forall s a P h A,
    h ||- P a ** fas P s ** A -> ~ set_in a s ->
      h ||- fas P (set_add a s) ** A.
  Proof.
    intros.
    apply (sconj_mono_left (sconj_assoc_right H)); intros.
    apply fas_intro1; trivial.
  Qed.

  Lemma fas_impl : forall s (P P' : nat -> hpred) h,
    h ||- fas P s -> (forall a, P a ==> P' a) ->
      h ||- fas P' s.
  Proof.
    induction s.
    induction s; intros.
    apply H.
    unfold fas in * ;simpl in * .
    eapply sconj_mono; intros.
    apply H.
    auto.
    destruct (andb_true_eq _ _ (sym_eq e)) as [B1 B2].
    eapply IHs; auto.
    apply H1.
    auto.
  Qed.

  Lemma fas_impl1 : forall s (P P' : nat -> hpred) h,
    h ||- fas P s -> (forall a, set_in a s -> P a ==> P' a) ->
      h ||- fas P' s.
  Proof.
    induction s.
    induction s; intros.
    apply H.
    unfold fas in * ;simpl in * .
    eapply sconj_mono; intros.
    apply H.
    apply H0; auto.
    left; trivial.
    destruct (andb_true_eq _ _ (sym_eq e)) as [B1 B2].
    eapply (IHs (sym_eq B2)); auto.
    apply H1.
    intros; apply H0; auto.
    right; apply H2.
  Qed.

End Fas.

Add Morphism fas : fas_ext.
  intros.
  rewrite <- (set_ext H).
  apply H0.
Qed.


(** ******************************************************************* *)
(** tactics *)
(** ******************************************************************* *)
Ltac sepelim Hp :=
  match type of Hp with
    | _ ||- unit => hypreplace Hp (unit_elim Hp); subst h
    | _ ||- bang _ => 
      let A := fresh "A" in let Hp' := fresh "Hp" in
	(destruct (bang_elim Hp) as  [Hp' A]; 
	  hypreplace Hp' (unit_elim Hp'); subst h; clear Hp')
    | _ ||- lex _ => 
      let x := fresh "x" in let Hp' := fresh "Hp'" in
	(destruct (lex_elim Hp) as [x Hp']; 
	  clear Hp; rename Hp' into Hp; sepelim Hp)

    | _ ||- sconj (bang _) _ => 
      let A := fresh "A" in let Hp' := fresh "Hp" in
	(destruct (bang_elim_left Hp) as  [A Hp']; 
	  clear Hp; rename Hp' into Hp; sepelim Hp)
    | _ ||- sconj _ (bang _) => 
      let A := fresh "A" in let Hp' := fresh "Hp" in
	(destruct (bang_elim_right Hp) as  [A Hp']; 
	  clear Hp; rename Hp' into Hp; sepelim Hp)

    | _ ||- sconj unit _ => hypreplace Hp (unit_elim_left Hp); sepelim Hp
    | _ ||- sconj _ unit => hypreplace Hp (unit_elim_right Hp); sepelim Hp
    | _ ||- sconj (sconj _ _) _ => hypreplace Hp (sconj_assoc_left Hp); sepelim Hp

    | _ ||- sconj _ _ =>  
      let h1 := fresh "h" in let h2 := fresh "h" in let J := fresh "J" in 
	let Hp1 := fresh "Hp" in let Hp2 := fresh "Hp" in
	  (destruct Hp as [h1 [h2 [J [Hp1 Hp2]]]];
	    sepelim Hp1; sepelim Hp2;
	    assert (Hp := sconj_intro Hp1 Hp2 J);
	      clear h1 h2 J Hp1 Hp2)
    | _ => idtac
  end.

Ltac assocl H :=
  repeat (hypreplace H (sconj_assoc_left H)).

Ltac assocr :=
  repeat (apply sconj_assoc_right).

Ltac matchhp ig ih sg sh :=
  match goal with
    | |- _ ||- top => apply top_intro (* true *)
    | [Hp : ?h ||- ?A |- ?h ||- ?A ] => exact Hp (* assumption *)
      
    | |- _ ||- unit => try (apply unit_intro)
    | |- _ ||- bang _ => apply bang_intro; auto

    | |- _ ||- sconj _ unit => apply unit_intro_right; 
      matchhp (pred ig) sh (pred sg) sh
    | |- _ ||- sconj unit _ => apply unit_intro_left;
      matchhp (pred ig) sh (pred sg) sh

    | |- _ ||- sconj (bang _) _ => apply bang_intro_left; 
      [auto | matchhp (pred ig) sh (pred sg) sh]
    | |- _ ||- sconj _ (bang _) => apply bang_intro_right; 
      [auto | matchhp (pred ig) sh (pred sg) sh]

    | [Hp : ?h ||- sconj top _ |- ?h ||- _ ] => (* top in assumption *)
      hypreplace Hp (top_elim_left Hp);
      let htop := fresh "htop" in
	(destruct Hp as [htop Hp]; matchhp ig ih sg sh)

    | [Hp : ?h ||- sconj _ top |- ?h ||- _ ] => (* top in assumption *)
      hypreplace Hp (top_elim_right Hp);
      let htop := fresh "htop" in
	(destruct Hp as [htop Hp]; matchhp ig ih sg sh)

    | [NI : ~ set_in ?x ?s |-
      ?h ||- fas ?P (set_add ?x ?s)] => apply fas_intro1; 
    [assocr; matchhp (S ig) sh (S sg) sh | assumption]

    | [NI : ~ set_in ?x ?s |-
      ?h ||- sconj (fas ?P (set_add ?x ?s)) _] => apply fas_intro; 
	[assocr; matchhp (S ig) sh (S sg) sh | assumption]
      
    | [Hp : ?h ||- sconj ?A ?B |- ?h ||- sconj ?A ?C] => (* match left *)
      let h1 := fresh "h" in
	(apply (sconj_mono_right Hp);
	  (clear h Hp; intros h Hp) || (clear Hp; intros h1 Hp);
	    assocl Hp; assocr;
            matchhp (pred ig) (pred sh) (pred sg) (pred sh))
    | [Hp : ?h ||- sconj ?A ?B |- ?h ||- sconj ?C ?B] => (* match right *)
      let h' := fresh "h" in
	(apply (sconj_mono_left Hp);
	  (clear h Hp; intros h Hp) || (clear Hp; intros h' Hp);
	    assocl Hp; assocr;
            matchhp (pred ig) (pred sh) (pred sg) (pred sh))
	
    | [Hp : ?h ||- sconj ?A ?B |- ?h ||- sconj ?C ?D] => (* shift *)
      match eval compute in ig with
	| O => (* no more shift on goal *)
	  match eval compute in ih with
	    | O => idtac (* finished *)
	    | _ => (* shift hyp, continue with matching *)
              hypreplace Hp (sconj_sym Hp); assocl Hp;
	      matchhp sg (pred ih) sg sh
	  end
	| _ => (* shift goal *)
	  apply sconj_sym; assocr; matchhp (pred ig) ih sg sh
      end
    | _ => idtac
  end.

Ltac simplsep sizeg sizeh :=
 match goal with
 | [Hp : ?h ||- _ |- ?h ||- _ ] =>
   assocl Hp; assocr; matchhp sizeg sizeh sizeg sizeh
 end.

Ltac lexintro size :=
  match goal with
    | |- _ ||- lex _ => apply lex_intro1
    | |- _ ||- sconj (lex _) _ => apply lex_intro_left; apply lex_intro1
    | |- _ ||- sconj _ (lex _) => apply lex_intro_right; apply lex_intro1
    | |- _ ||- sconj _ _ => 
      match eval compute in size with
	| O => idtac
	| S _ => assocr; apply sconj_sym; assocr; lexintro (pred size)
      end
  end.

Ltac hvinject' i1 i2 s1 s2 :=
  match goal with 
    | [ Hp1 : ?h ||- (?addr |-> ?v1), 
	Hp2 : ?h ||- (?addr |-> ?v2) |- _] =>
    let Eq := fresh "Eq" in
      assert (Eq := cont_cont_eq Hp1 Hp2); 
	try (clear Hp1 Hp2 h)
    | [ Hp1 : ?h ||- sconj (?addr |-> ?v1) _ , 
	Hp2 : ?h ||- sconj (?addr |-> ?v2) _ |- _] =>
    let Eq := fresh "Eq" in 
      let h11 := fresh "h" in let h12 := fresh "h" in 
	let h21 := fresh "h" in let h22 := fresh "h" in 
	  let J1 := fresh "J" in let J2 := fresh "J" in 
	    let Hp11 := fresh "Hp" in let Hp12 := fresh "Hp" in 
	      let Hp21 := fresh "Hp" in let Hp22 := fresh "Hp" in 
		(assert (Eq := cont_cont_eq1 Hp1 Hp2);
		  destruct Hp1 as [h11 [h12 [J1 [Hp11 Hp12]]]];
		    destruct Hp2 as [h21 [h22 [J2 [Hp21 Hp22]]]];
		      rewrite Eq in Hp11;
 			try (rewrite (join_sing_eq Hp11 J1 Hp21 J2) in Hp12;
			  clear h11 h21 h12 J1 J2 Hp11 Hp21; try (clear h);
			    assocl Hp12; assocl Hp22; 
			    hvinject' (pred i1) (pred i2) (pred s1) (pred s2)))
    | [ Hp1 : ?h ||- sconj _ _, 
	Hp2 : ?h ||- sconj _ _ |- _] =>
    match eval compute in i1 with
      | O => (* no more shift on Hp1 *)
	match eval compute in i2 with
	  | O => try (clear Hp1 Hp2); try (clear h)(* finished *)
	  | _ => (* shift hyp, continue with matching *)
            hypreplace Hp2 (sconj_sym Hp2); assocl Hp2;
	    hvinject' s1 (pred i2) s1 s2
	end
      | _ => (* shift on Hp2 *)
        hypreplace Hp1 (sconj_sym Hp1); assocl Hp1;
	hvinject' (pred i1) i2 s1 s2
    end
  end.

Ltac hvinject s1 s2 :=
  match goal with 
    | [ Hp1 : ?h ||- _, Hp2 : ?h ||- _ |- _] =>
      generalize Hp1; generalize Hp2;
	assocl Hp1; assocl Hp2; 
	hvinject' s1 s2 s1 s2; intros
  end.

Ltac sconjlift n :=
  match eval compute in n with
    | O => assocr
    | _ => assocr; apply sconj_sym; sconjlift (pred n)
  end.

Ltac sconjlifth n H :=
  match eval compute in n with
    | O => assocl H
    | _ => assocl H; hypreplace H (sconj_sym H); sconjlifth (pred n) H
  end.


Lemma heap_eq_add : forall h x v h',
  h ||- x |-> v ** hp_eq h' ->
    h ||- hp_eq (heap_set h' x v).
Proof.
  intros h x v h' Hp.
  destruct Hp as [h1 [h2 [J [Hp1 Hp2]]]].
  unfold hp_eq in * ; unfold hpred_true in * .
  destruct Hp1 as [Hptr Hv].
  unfold singleton_heap in Hv.
  apply heap_ext; intros l.
  compare l x; intro e; subst.
  generalize (J x); simplhp; trivial; intros.
  destruct (heap_get h x); destruct (heap_get h' x); 
    simpl in H; subst; try tauto.
  generalize (J l); simplhp; trivial; intros.
  destruct (heap_get h l); destruct (heap_get h' l);
    try (simpl in H; subst; tauto).
Qed.

Lemma heap_eq_add1 : forall h x v h' A,
  h ||- x |-> v ** hp_eq h' ** A ->
    h ||- hp_eq (heap_set h' x v) ** A.
Proof.
  intros; simplsep 2 2; apply heap_eq_add; trivial.
Qed.

Lemma cont_intro2 : forall h x v,
  heap_get h x = Some v -> h ||- x |-> v ** top.
Proof.
  intros h x v E.
  assert (Px := heap_get_ptr E).
  assert (heap_subheap (singleton_heap x v) h).
  intros a v'.
  unfold singleton_heap.
  compare x a; intro e; subst; simplhp; trivial; intros.
  injection H; intros; subst; trivial.
  discriminate H.
  destruct (heap_subheap_ex H) as [hb Hb].
  apply sconj_intro with (singleton_heap x v) hb; trivial;
    unfold cont; unfold hpred_true; unfold top; auto.
Qed.

Lemma heap_add_ok : forall h A x v,
  h ||- A -> ptr x ->
    (forall ff, ~ (h ||- x |-> ff ** top)) -> 
    heap_set h x v ||- x |-> v ** A.
Proof.
  intros h A x v Hp Hptr FF.
  apply sconj_intro with (ha := heap_set empty_heap x v)
    (hb := h); trivial.
  unfold hpred_true; unfold cont; split; trivial.
  assert (N : heap_get h x = None).
  destruct (heap_get_dec2 h x) as [| [vv Hvv]]; trivial.
  hypreplace Hvv (cont_intro2 Hvv); firstorder.
  intro l.
  compare l x; intro e; subst; simplhp; trivial.
  rewrite N; constructor.
  destruct (heap_get h l); constructor.
Qed.
