(** Basic machine model *)

(** machine state *)

(** Chunxiao Lin @ 2006 *)

Require Export heap2.
Require Export rfile.
Require Import natutil.

Section State.
  
  Definition state := (prod rfile heap).

  Definition rf_of (s : state) := fst s.

  Definition hp_of (s : state) := snd s.

  Definition rget (s : state) r :=
    rfile_get (rf_of s) r.

  Definition rset (s : state) r val :=
    (rfile_set (rf_of s) r val, hp_of s).

  Definition hget (s : state) addr :=
    heap_get (hp_of s) addr.

  (** always success *)
  Definition hset (s : state) addr val  :=
    (rf_of s, heap_set (hp_of s) addr val).

  (** write to non-exist addr will go stuct *)
  Definition hset' (s : state) addr val  :=
    match hget s addr with
      | Some v => Some (hset s addr val)
      | None => None
    end.

End State.

Section StateTheory.

  Variable s : state.
  Variable addr addr' val : nat.
  Variable r r' : reg.

  Hint Resolve rfile_set_get_eq rfile_set_get_neq
    heap_set_get_eq heap_set_get_ne : st_theory.
  
  Ltac solvest := 
    intros;
    try unfold rset; try unfold rget;
      try unfold hset; try unfold hget;
	try unfold rf_of; try unfold hp_of;
	  destruct s as [rf hp]; simpl; 
	    auto with st_theory.

  Lemma rset_rget_eq : r <> r0 -> rget (rset s r val) r = val.
  Proof.
    solvest.
  Qed.

  Lemma rget_zero : rget s r0 = 0.
  Proof.
    solvest.
  Qed.

  Lemma rset_rget_neq : 
    r <> r' ->
    rget (rset s r' val) r = rget s r.
  Proof.
    solvest.
  Qed.

  Lemma hset_hget_eq : ptr addr -> hget (hset s addr val) addr = Some val.
  Proof.
    solvest.
  Qed.

  Lemma hset_hget_neq : 
    addr <> addr' ->
    hget (hset s addr' val) addr = hget s addr.
  Proof.
    solvest.
  Qed.

  Lemma rset_hp_eq :
    hp_of (rset s r val) = hp_of s.
  Proof.
    auto.
  Qed.

  Lemma hset_rf_eq :
    rf_of (hset s addr val) = rf_of s.
  Proof.
    auto.
  Qed.

  Lemma rset_hget_eq :
    hget (rset s r val) addr = hget s addr.
  Proof.
    solvest.
  Qed.

  Lemma hset_rget_eq :
    rget (hset s addr val) r = rget s r.
  Proof.
    solvest.
  Qed.

End StateTheory.

Lemma p0id : forall x,
  x + 0 = x.
Proof.
  intros; auto with arith.
Qed.

Lemma pm2 : forall x,
  x + 2 - 2 = x.
Proof.
  intros; omega.
Qed.

Lemma mp2 : forall x,
  x >= 2 -> x - 2 + 2 = x.
Proof.
  intros; omega.
Qed.

Lemma p2m1 : forall x,
  x + 2 - 1 = S x.
Proof.
  intros; omega.
Qed.

Lemma p1S : forall x,
  x + 1 = S x.
Proof.
  intros; omega.
Qed.

Lemma pSSp : forall x y,
  x + S y = S x + y.
Proof.
  intros; omega.
Qed.

Lemma SSm2 : forall x,
  (S (S x)) - 2 = x.
Proof.
  intros; omega.
Qed.

Lemma SSm1 : forall x,
  (S (S x)) - 1 = S x.
Proof.
  intros; omega.
Qed.

Lemma pm : forall n m, n + m - m = n.
Proof.
  intros; omega.
Qed.

Ltac simplpm :=
  match goal with
    | [ |- context [(?x + ?y - ?y)]] =>
      rewrite pm; simplpm
    | [ |- context [(?x + 1)]] =>
      rewrite p1S; simplpm
    | [ |- context [(?x + 0)]] =>
      rewrite p0id; simplpm
    | _ => idtac
  end.

Ltac simplpm2 :=
  match goal with
    | [ |- context [(?x + 2 - 2)]] =>
      rewrite pm2; simplpm2
    | [ |- context [(?x - 2 + 2)]] =>
      rewrite mp2;
	[simplpm2 | auto]
    | [ |- context [(?x + 2 - 1)]] =>
      rewrite p2m1; simplpm2
    | _ => simplpm
  end.

Ltac simplpm2h H :=
  generalize H; clear H; simplpm2; intro H.

Ltac simplst := 
  match goal with
    | [ H : ptr ?a |- context [(hget (hset ?s ?a ?v) ?a)] ] => 
      rewrite hset_hget_eq; [simplst | auto]
    | [ |- context [(hget (hset ?s ?a2 ?v) ?a1)] ] => 
      rewrite hset_hget_neq; [simplst | auto]

    | [ |- context [(rget ?s r0)] ] => 
      rewrite rget_zero; simplst
    | [ |- context [(rget (rset ?s ?r ?v) ?r)] ] => 
      rewrite rset_rget_eq; 
	[simplst | let F := fresh "F" in (intro F; discriminate F)]
    | [ |- context [(rget (rset ?s ?r2 ?v) ?r1)] ] => 
      rewrite rset_rget_neq;
	[simplst | auto; discriminate]

    | [ |- context [(rget (hset ?s ?a ?v) ?r)] ] => 
      rewrite hset_rget_eq; simplst
    | [ |- context [(hget (rset ?s ?r ?v) ?a)] ] => 
      rewrite rset_hget_eq; simplst

    | [ |- context [(rf_of (hset ?s ?a ?v))] ] => 
      rewrite hset_rf_eq; simplst
    | [ |- context [(hp_of (rset ?s ?r ?v))] ] => 
      rewrite rset_hp_eq; simplst

    | _ => idtac
  end.

Hint Rewrite p0id pm2 mp2 p2m1 p1S pSSp SSm2 SSm1 : pm.
Hint Rewrite hset_hget_eq hset_hget_neq rget_zero 
  rset_rget_neq rset_rget_eq  hset_rget_eq
  rset_hget_eq hset_rf_eq rset_hp_eq : st.

Ltac strewrite := autorewrite with st; autorewrite with pm; auto; try discriminate.